#include "LevApproximation.h"
#include <fstream>
#include <iostream>
#include <algorithm>

float influenceFunction(float selfOpinion,float opinionOther,float sigma){
  return 1.0/(1.0+exp((selfOpinion-opinionOther)/sigma));
}

float dInfluenceFunction(float selfOpinion,float opinionOther,float sigma){
  float h= influenceFunction(selfOpinion,opinionOther,sigma);
  return (-h*(1-h))/sigma;
}

LevApproximation::LevApproximation(float delta_,float sigma_,unsigned int nbAgents_,unsigned int nbGossips_,float rangeInitOpinions_):
  delta(delta_),sigma(sigma_),gossip(nbGossips_),nbStep(0),startReputation(nbAgents_,0),opinions(200,nbAgents_,nbAgents_,0),avgMulti(nbAgents_,nbAgents_,nbAgents_,nbAgents_,0),
  influences(nbAgents_,nbAgents_,0),dInfluences(nbAgents_,nbAgents_,0),deltaOpinions(nbAgents_,nbAgents_,0),delta2Sur3(pow(delta_,2)/3.0),equi(200,nbAgents_,0)
{
  float tmpOpinion=-rangeInitOpinions_/2.0;
  for(IdAgent a1=0;a1<nbAgents_;a1++){
    startReputation[a1]=tmpOpinion;
    tmpOpinion+=rangeInitOpinions_/(nbAgents_-1);
  }
  updateInflus();
}

void LevApproximation::updateInflus(){
  for(unsigned int i=0;i<getNbAgents();i++){
    for(unsigned int j=0;j<getNbAgents();j++){
      influences[i][j]=influenceFunction(startReputation[i]+opinions[nbStep][i][i],startReputation[j]+opinions[nbStep][i][j],sigma);
      dInfluences[i][j]=dInfluenceFunction(startReputation[i]+opinions[nbStep][i][i],startReputation[j]+opinions[nbStep][i][j],sigma);
    }
  }
}

float LevApproximation::z(IdAgent i, IdAgent j){
  return opinions[nbStep][i][i]-opinions[nbStep][i][j];
}

float LevApproximation::influCircum(IdAgent i,IdAgent j){
  return influences[i][j]-dInfluences[i][j]*z(i,j);
}

float LevApproximation::F(IdAgent i,IdAgent j){
  return opinions[nbStep][i][i]+influCircum(i,j)*(opinions[nbStep][j][i]-opinions[nbStep][i][i])-dInfluences[i][j]*(avgMulti[i][i][i][i]-avgMulti[i][i][j][i]);
}

float LevApproximation::F2(IdAgent i,IdAgent j){
  return
  avgMulti[i][i][i][i]
  +pow(influCircum(i,j),2)*(avgMulti[j][i][j][i]+avgMulti[i][i][i][i]-2*avgMulti[j][i][i][i])
  +2*influCircum(i,j)*(avgMulti[i][i][j][i]-avgMulti[i][i][i][i]);
}

float LevApproximation::Fx(IdAgent i,IdAgent j,IdAgent k,IdAgent l){
  return avgMulti[i][i][k][l]+influCircum(i,j)*(avgMulti[j][i][k][l]-avgMulti[i][i][k][l]);
}

float LevApproximation::G(IdAgent j,IdAgent i){
  return opinions[nbStep][j][i]+influCircum(j,i)*(opinions[nbStep][i][i]-opinions[nbStep][j][i])+dInfluences[j][i]*(avgMulti[j][i][j][i]-avgMulti[i][i][j][i]);
}

float LevApproximation::G2(IdAgent j,IdAgent i){
  return
  avgMulti[j][i][j][i]
  +pow(influCircum(j,i),2)*(avgMulti[i][i][i][i]+avgMulti[j][i][j][i]-2*avgMulti[i][i][j][i])
  +2*influCircum(j,i)*(avgMulti[j][i][i][i]-avgMulti[j][i][j][i]);
}

float LevApproximation::Gx(IdAgent j,IdAgent i,IdAgent k,IdAgent l){
  return avgMulti[j][i][k][l]+influCircum(j,i)*(avgMulti[i][i][k][l]-avgMulti[j][i][k][l]);
}

float LevApproximation::J(IdAgent i,IdAgent j,IdAgent g){
  return opinions[nbStep][i][g]+influCircum(i,j)*(opinions[nbStep][j][g]-opinions[nbStep][i][g]);
}

float LevApproximation::J2(IdAgent i,IdAgent j,IdAgent g){
  return
  avgMulti[i][g][i][g]
  +pow(influCircum(i,j),2)*(avgMulti[j][g][j][g]+avgMulti[i][g][i][g]-2*avgMulti[j][g][i][g])
  +2*influCircum(i,j)*(avgMulti[j][g][i][g]-avgMulti[i][g][i][g]);
}

float LevApproximation::Jx(IdAgent i,IdAgent j,IdAgent g,IdAgent k,IdAgent l){
  return avgMulti[i][g][k][l]+influCircum(i,j)*(avgMulti[j][g][k][l]-avgMulti[i][g][k][l]);
}

float LevApproximation::JJ(IdAgent i,IdAgent j,IdAgent g,IdAgent k,IdAgent l,IdAgent m){
  return
  avgMulti[i][g][k][m]
  +influCircum(i,j)*influCircum(k,l)*(avgMulti[j][g][l][m]-avgMulti[j][g][k][m]-avgMulti[i][g][l][m]+avgMulti[i][g][k][m])
  +influCircum(k,l)*(avgMulti[l][m][i][g]-avgMulti[k][m][i][g])
  +influCircum(i,j)*(avgMulti[j][g][k][m]-avgMulti[i][g][k][m]);
}

float LevApproximation::FG_ijji(IdAgent i,IdAgent j){
  return
  avgMulti[i][i][j][i]
  +influCircum(i,j)*influCircum(j,i)*(avgMulti[j][i][i][i]-avgMulti[j][i][j][i]-avgMulti[i][i][i][i]+avgMulti[i][i][j][i])
  +influCircum(i,j)*(avgMulti[j][i][j][i]-avgMulti[j][i][i][i])
  +influCircum(j,i)*(avgMulti[i][i][i][i]-avgMulti[i][i][j][i]);
}

void LevApproximation::averageInteraction(){
  float Na=getNbAgents();
  float Nc=(Na-1)*Na;
  float Nt=Na*(Na-1)*(Na-2);
  float P1=1-2/Nc-(2*gossip)/Nc;
  float P2=1-2/Na-(2*gossip)/Nc;
  float P3=1-4/Nc-(2*gossip)/Nt-4*gossip*(Na-3)/Nt;
  float sii,sii2,siiji,sji,sjiji,siiji2,sjipi;
  float st1,st2,st3,st4,st5;
  NVector<4, float> newAvgMulti(Na,Na,Na,Na,0);
  //extend the size of the vector if it is necessary (or close to be)
  if((nbStep+5)>opinions.size()){
    opinions.push_back(NVector<2,float>(Na,Na,0));
    equi.push_back(NVector<1,float>(Na,0));
  }
  for(IdAgent i=0;i<Na;i++){
    sii=0;sii2=0;st1=0;st2=0;st3=0;st4=0;st5=0;
    for(IdAgent j=0;j<Na;j++){
      if(j!=i){
        st1+=influCircum(i,j)/influCircum(j,i);
        st2+=dInfluences[i][j]*(avgMulti[i][i][i][i]-avgMulti[i][i][j][i]);
        st3+=influCircum(i,j)/influCircum(j,i)*dInfluences[j][i]*(avgMulti[j][i][j][i]-avgMulti[i][i][j][i]);
        sii+=F(i,j);
        sii2+=F2(i,j)+pow(influences[i][j],2)*pow(delta,2)/3;
        siiji=0;sji=0;sjiji=0;siiji2=0;
        for(IdAgent p=0;p<Na;p++){
          if(p!=i && p!=j){
            st5+=influCircum(i,j)/influCircum(j,i)*influCircum(j,p)*(opinions[nbStep][p][i]-opinions[nbStep][j][i]);
            sji+=J(j,p,i);
            sjiji+=J2(j,p,i)+pow(influences[j][p],2)*pow(delta,2)/3;
            siiji+=Fx(i,p,j,i);
            siiji2+=Jx(j,p,i,i,i);
          }
        }
        opinions[nbStep+1][j][i]=P1*opinions[nbStep][j][i]+2/Nc*G(j,i)+(2*gossip/Nt)*sji;
        newAvgMulti[j][i][j][i]=P1*avgMulti[j][i][j][i] + 2/Nc*(G2(j,i)+pow(influences[j][i],2)*pow(delta,2)/3)
          +(2*gossip/Nt)*sjiji;
        newAvgMulti[i][i][j][i]=P2*avgMulti[i][i][j][i] + 2/Nc*(FG_ijji(i,j)+siiji)
          +(2*gossip/Nt)*siiji2;

        st4+=influCircum(i,j)/influCircum(j,i)*opinions[nbStep+1][j][i];
      }
    }
    opinions[nbStep+1][i][i]=(Na-2.)/Na*opinions[nbStep][i][i]+2./Nc*sii;
    newAvgMulti[i][i][i][i]=(Na-2.)/Na*avgMulti[i][i][i][i]+2./Nc*sii2;
    equi[nbStep+1][i]=equi[nbStep][i]+2./(Nc*(1+st1))*(-st2+st3+gossip*st5/(Na-2.));
  }
  for(IdAgent i=0;i<Na;i++){
    for(IdAgent p=0;p<Na;p++){
      for(IdAgent j=0;j<Na;j++){
        if(i!=p && p!=j && i!=j){
          sjipi=0;
          for(IdAgent q=0;q<Na;q++){
            if(q!=j && q!=i && q!=p)
              sjipi+=Jx(j,q,i,p,i)+Jx(p,q,i,j,i);
          }
          newAvgMulti[j][i][p][i]=P3*avgMulti[j][i][p][i]+2/Nc*(Gx(p,i,j,i)+Gx(j,i,p,i))
            +(2*gossip/Nt)*JJ(j,p,i,p,j,i)+(2*gossip/Nt)*sjipi;
        }
      }
    }
  }

  for(unsigned int i=0;i<Na;i++){
    for(unsigned int j=0;j<Na;j++){
      for(unsigned int k=0;k<Na;k++){
        for(unsigned int l=0;l<Na;l++){
          if(newAvgMulti[i][j][k][l]==0 && newAvgMulti[k][l][i][j]==0){
//            std::cout<<"Les deux null "<<i<<" "<<j<<" "<<k<<" "<<l<<std::endl;
          }else if(newAvgMulti[i][j][k][l]!=0 && newAvgMulti[k][l][i][j]!=0 && newAvgMulti[i][j][k][l]!=newAvgMulti[k][l][i][j]){
//            std::cout<<"les deux different "<<i<<" "<<j<<" "<<k<<" "<<l<<" "<<newAvgMulti[i][j][k][l]<<" "<<newAvgMulti[k][l][i][j]<<std::endl;
          }else if(newAvgMulti[i][j][k][l]!=0 && newAvgMulti[k][l][i][j]==0){
            newAvgMulti[k][l][i][j]=newAvgMulti[i][j][k][l];
          }else if(newAvgMulti[i][j][k][l]==0 && newAvgMulti[k][l][i][j]!=0){
            newAvgMulti[i][j][k][l]=newAvgMulti[k][l][i][j];
          }
        }
      }
    }
  }
  avgMulti=newAvgMulti;
  nbStep++;
  updateInflus();
}

void LevApproximation::writeOpinions(std::string fileName){
  std::ofstream myfile;
  myfile.open(fileName);
  myfile<<"step,rep(i),rep(j),xii,x_ji,equilibrium"<<std::endl;

  for(unsigned int t=0;t<nbStep;t++){
    for(IdAgent i=0;i<getNbAgents();i++){
      for(IdAgent j=0;j<getNbAgents();j++){
        if(i!=j){
          myfile<<t<<","<<startReputation[i]<<","<<startReputation[j]<<",";
          myfile<<opinions[t][i][i]<<","<<opinions[t][j][i]<<","<<equi[t][i]<<std::endl;
        }
      }
    }
  }
  myfile.close();
}
