#include "Leviathan.h"

#include <fstream>
#include <chrono>

//-----------------------------------------------------------------Utilitary

float average(const std::vector<float> & tab){
  float ret=0;
  for(unsigned int i=0;i<tab.size();i++)
    ret+=tab[i];
  return ret/(float)tab.size();
}

float standardDeviation(const std::vector<float> & tab){
  float avg=average(tab);
  float ret=0;
  if(tab.size()!=0){
    for(unsigned int i=0;i<tab.size();i++)
      ret+=pow(tab[i]-avg,2.);
    ret=sqrt(ret/(float)(tab.size()));
  }
  return ret;
}

std::mt19937 g_m_t_generator(std::chrono::system_clock::now().time_since_epoch().count());

//-----------------------------------------------------------------Agent

void Agent::init(unsigned int nbAgents){
  opinions=std::vector<float>(nbAgents,0);
  deltaOpinions=std::vector<float>(nbAgents,0);
  for(IdAgent i=0;i<nbAgents;i++){
    if(i!=id)
      gossipAbout.push_back(i);
  }
}

IdAgent Agent::gossipToWho(){
  IdAgent Aq;
  Aq=0;
  if(countGossipAbout>=gossipAbout.size())
    Aq=UINT_MAX;
  else{
    unsigned int pos;
    do{
      pos = std::uniform_int_distribution<int>(countGossipAbout, gossipAbout.size()-1)(g_m_t_generator);
      Aq=gossipAbout[pos];
      std::swap(gossipAbout[countGossipAbout],gossipAbout[pos]);
      countGossipAbout++;
    }while(Aq==directInteraction && countGossipAbout<gossipAbout.size());
    if(Aq==directInteraction)
      Aq=UINT_MAX;
    dady->agents[directInteraction].listGossip.push_back(Aq);
  }
  return Aq;
}

void Agent::updateOpinions(){
  opinions[directInteraction]=dady->borne(opinions[directInteraction]+deltaOpinions[directInteraction]);
  deltaOpinions[directInteraction]=0;
  opinions[id]=dady->borne(opinions[id]+deltaOpinions[id]);
  deltaOpinions[id]=0;
  for(unsigned int i=0;i<listGossip.size();i++){
    opinions[listGossip[i]]=dady->borne(opinions[listGossip[i]]+deltaOpinions[listGossip[i]]);
    deltaOpinions[listGossip[i]]=0;
  }
  listGossip.clear();
  countGossipAbout=0;
}

//------------------------------------------------------------------Leviathan

float Leviathan::borne(float value){
  float ret=value;
  if(useBorder){
    if(value>1)
      ret= 1;
    else if (value<-1)
      ret= -1;
  }
  return ret;
}

Leviathan::Leviathan(float delta_,float sigma_,unsigned int nbAgents_,unsigned int nbGossips_):
  delta(delta_),gossip(nbGossips_),sigma(sigma_),nbStep(0),oneInteractionParStep(false),g_noise(-delta_, delta_)
{
  for(unsigned int i=0;i<nbAgents_;i++){
    agents.push_back(Agent(this,i));
    agents.back().init(nbAgents_);
  }
}

void Leviathan::step(){
  IdAgent Ai;
  IdAgent Aj;
  float nbInteract;
  if(oneInteractionParStep==true)
      nbInteract=1;
  else
      nbInteract=getNbAgents();
  for(unsigned int i=0;i<nbInteract;i++){
    selectCoupleForInterraction(Ai,Aj);
    directInteraction(Ai,Aj);
    directInteraction(Aj,Ai);
    gossipInteraction(Ai,Aj);
    agents[Ai].updateOpinions();
    agents[Aj].updateOpinions();
  }
  nbStep ++;
}

void Leviathan::selectCoupleForInterraction(IdAgent & Ai,IdAgent & Aj){
  Ai=getRandomAgent();
  do{
    Aj=getRandomAgent();
  }while(Ai==Aj);
  agents[Ai].directInteraction=Aj;
  agents[Aj].directInteraction=Ai;
}

void Leviathan::directInteraction(IdAgent Ai,IdAgent Aj){
  float p = 1. / (1. + exp((getOpinion(Ai,Ai)-getOpinion(Ai,Aj))/sigma));
  agents[Ai].influenceOther = p;
  agents[Ai].deltaOpinions[Ai] += p * (getOpinion(Aj,Ai) - getOpinion(Ai,Ai) + g_noise(g_m_t_generator));
  agents[Ai].deltaOpinions[Aj] += p * (getOpinion(Aj,Aj) - getOpinion(Ai,Aj) + g_noise(g_m_t_generator));
}

void Leviathan::gossipInteraction(IdAgent Ai,IdAgent Aj){
  for(unsigned int g=0;g<gossip;g++){
    oneGossip(Ai,Aj);
    oneGossip(Aj,Ai);
  }
}

void Leviathan::oneGossip(IdAgent Ai,IdAgent Aj){
  IdAgent Aq=agents[Aj].gossipToWho();
  if(Aq!=UINT_MAX){
    float p=agents[Ai].influenceOther;
    agents[Ai].deltaOpinions[Aq] += p * (getOpinion(Aj,Aq) - getOpinion(Ai,Aq)+ g_noise(g_m_t_generator) );
  }
}

//------------------------------------------------------------------Leviathan - Setters

void Leviathan::setAgentReputation(IdAgent Ai,float newReputation){
  for(IdAgent i=0;i<getNbAgents();i++)
    agents[i].opinions[Ai]=newReputation;
}

//------------------------------------------------------------------Leviathan - Getters

float Leviathan::getAverageOpinion(){
  float sum=0;
  for(IdAgent i=0;i<getNbAgents();i++){
    for(IdAgent j=0;j<getNbAgents();j++){
      sum+=getOpinion(i,j);
    }
  }
  return sum/(getNbAgents()*getNbAgents());
}

float Leviathan::getAverageOpinionTo(IdAgent a){
  float ret=0;
  for(IdAgent i=0;i<getNbAgents();i++){
    ret += getOpinion(i,a);
  }
  return ret /getNbAgents();
}

float Leviathan::getStdOpinion(){
  std::vector<float> ops;
  for(IdAgent i=0;i<getNbAgents();i++){
    for(IdAgent j=0;j<getNbAgents();j++)
      ops.push_back(getOpinion(i,j));
  }
  return standardDeviation(ops);
}

float Leviathan::getStdOpinionTo(IdAgent a){
  std::vector<float> ops;
  for(IdAgent i=0;i<getNbAgents();i++){
    ops.push_back(getOpinion(i,a));
  }
  return standardDeviation(ops);
}

float Leviathan::getAverageStdOpinion(){
  float ret=0;
  for(IdAgent a=0;a<getNbAgents();a++)
    ret+=getStdOpinionTo(a);
  return ret/getNbAgents();
}

float Leviathan::getStdReputation(){
  std::vector<float> reputations;
  for(IdAgent a=0;a<getNbAgents();a++){
    reputations.push_back(getAverageOpinionTo(a));
  }
  return standardDeviation(reputations);
}

void Leviathan::writeOpinions(std::string fileName){
  std::ofstream myfile;
  myfile.open(fileName);
  myfile << "Run,population,sigma,noise_intensity,gossip";
  for(IdAgent i=0;i<getNbAgents();i++)
    myfile<<","<<"OPI"<<i;
  myfile<<"\n";
  for(IdAgent i=0;i<getNbAgents();i++){
    myfile<<nbStep<<","<<getNbAgents()<<","<<sigma<<","<<delta<<","<<gossip;
    for(IdAgent j=0;j<getNbAgents();j++)
      myfile<<","<<getOpinion(i,j);
    myfile<<"\n";
  }
  myfile.close();
}
